@@ -1,5 +1,6 @@
 /* FreeTDS - Library of routines accessing Sybase and Microsoft databases
  * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005 Brian Bruns
+ * Copyright (C) 2015  Frediano Ziglio
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -17,22 +18,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#if HAVE_CONFIG_H
 #include <config.h>
-#endif
 
 #include <stdarg.h>
 
-#if TIME_WITH_SYS_TIME
-# include <sys/time.h>
-# include <time.h>
-#else
-# if HAVE_SYS_TIME_H
-#  include <sys/time.h>
-# else
-#  include <time.h>
-# endif
-#endif
+#include <freetds/time.h>
 
 #include <stdio.h>
 #include <assert.h>
@@ -47,12 +37,11 @@
 
 #include "cspublic.h"
 #include "ctlib.h"
-#include "tdsconvert.h"
+#include <freetds/convert.h>
 #include "replacements.h"
 
-TDS_RCSID(var, "$Id: cs.c,v 1.76 2010/10/05 08:36:36 freddy77 Exp $");
+#undef cs_dt_crack
 
-static int _cs_datatype_length(int dtype);
 static CS_INT cs_diag_storemsg(CS_CONTEXT *context, CS_CLIENTMSG *message);
 static CS_INT cs_diag_clearmsg(CS_CONTEXT *context, CS_INT type);
 static CS_INT cs_diag_getmsg(CS_CONTEXT *context, CS_INT idx, CS_CLIENTMSG *message);
@@ -93,43 +82,6 @@ cs_prretcode(int retcode)
 	return unknown;
 }
 
-/**
- * returns the fixed length of the specified data type, or 0 if not a 
- * fixed length data type
- */
-static int 
-_cs_datatype_length(int dtype)
-{
-	tdsdump_log(TDS_DBG_FUNC, "_cs_datatype_length(%d)\n", dtype);
-
-	switch (dtype) {
-		case SYBINT1:
-			return 1;
-		case SYBINT2:
-			return 2;
-		case SYBINT4:
-			return 4;
-		case SYBINT8:
-			return 8;
-		case SYBFLT8:
-			return 8;
-		case SYBREAL:
-			return 4;
-		case SYBBIT:
-			return 1;
-		case SYBMONEY:
-			return 8;
-		case SYBMONEY4:
-			return 4;
-		case SYBDATETIME:
-			return 8;
-		case SYBDATETIME4:
-			return 4;
-		default:
-			return 0;
-	}
-}
-
 static const char *
 _cs_get_layer(int layer)
 {
@@ -250,7 +202,7 @@ _cs_locale_alloc(void)
 {
 	tdsdump_log(TDS_DBG_FUNC, "_cs_locale_alloc()\n");
 
-	return (CS_LOCALE *) calloc(1, sizeof(CS_LOCALE));
+	return tds_new0(CS_LOCALE, 1);
 }
 
 static void
@@ -369,7 +321,7 @@ cs_ctx_alloc(CS_INT version, CS_CONTEXT ** ctx)
 
 	tdsdump_log(TDS_DBG_FUNC, "cs_ctx_alloc(%d, %p)\n", version, ctx);
 
-	*ctx = (CS_CONTEXT *) calloc(1, sizeof(CS_CONTEXT));
+	*ctx = tds_new0(CS_CONTEXT, 1);
 	tds_ctx = tds_alloc_context(*ctx);
 	if (!tds_ctx) {
 		free(*ctx);
@@ -461,7 +413,7 @@ cs_config(CS_CONTEXT * ctx, CS_INT action, CS_INT property, CS_VOID * buffer, CS
 			free(ctx->userdata);
 	
 			if (buflen == CS_NULLTERM) {
-				maxcp = strlen(buffer) + 1;
+				maxcp = strlen((char*) buffer) + 1;
 			} else {
 				maxcp = buflen;
 			}
@@ -513,7 +465,8 @@ cs_config(CS_CONTEXT * ctx, CS_INT action, CS_INT property, CS_VOID * buffer, CS
 CS_RETCODE
 cs_convert(CS_CONTEXT * ctx, CS_DATAFMT * srcfmt, CS_VOID * srcdata, CS_DATAFMT * destfmt, CS_VOID * destdata, CS_INT * resultlen)
 {
-	int src_type, src_len, desttype, destlen, len, i = 0;
+	TDS_SERVER_TYPE src_type, desttype;
+	int src_len, destlen, len, i = 0;
 	CONV_RESULT cres;
 	unsigned char *dest;
 	CS_RETCODE ret;
@@ -551,19 +504,25 @@ cs_convert(CS_CONTEXT * ctx, CS_DATAFMT * srcfmt, CS_VOID * srcdata, CS_DATAFMT
 
 	}
 
-	src_type = _ct_get_server_type(srcfmt->datatype);
+	src_type = _ct_get_server_type(NULL, srcfmt->datatype);
+	if (src_type == TDS_INVALID_TYPE)
+		return CS_FAIL;
 	src_len = srcfmt->maxlength;
 	if (srcfmt->datatype == CS_VARCHAR_TYPE || srcfmt->datatype == CS_VARBINARY_TYPE) {
 		CS_VARCHAR *vc = (CS_VARCHAR *) srcdata;
 		src_len = vc->len;
 		srcdata = vc->str;
 	}
-	desttype = _ct_get_server_type(destfmt->datatype);
+	desttype = _ct_get_server_type(NULL, destfmt->datatype);
+	if (desttype == TDS_INVALID_TYPE)
+		return CS_FAIL;
 	destlen = destfmt->maxlength;
 	if (destfmt->datatype == CS_VARCHAR_TYPE || destfmt->datatype == CS_VARBINARY_TYPE) {
 		destvc = (CS_VARCHAR *) destdata;
 		destlen  = sizeof(destvc->str);
 		destdata = destvc->str;
+	} else if (is_numeric_type(desttype)) {
+		destlen = sizeof(TDS_NUMERIC);
 	}
 
 	tdsdump_log(TDS_DBG_FUNC, "converting type %d (%d bytes) to type = %d (%d bytes)\n",
@@ -663,9 +622,13 @@ cs_convert(CS_CONTEXT * ctx, CS_DATAFMT * srcfmt, CS_VOID * srcdata, CS_DATAFMT
 			}
 			break;
 		case SYBINT1:
+		case SYBUINT1:
 		case SYBINT2:
+		case SYBUINT2:
 		case SYBINT4:
+		case SYBUINT4:
 		case SYBINT8:
+		case SYBUINT8:
 		case SYBFLT8:
 		case SYBREAL:
 		case SYBBIT:
@@ -673,8 +636,13 @@ cs_convert(CS_CONTEXT * ctx, CS_DATAFMT * srcfmt, CS_VOID * srcdata, CS_DATAFMT
 		case SYBMONEY4:
 		case SYBDATETIME:
 		case SYBDATETIME4:
-			*resultlen = _cs_datatype_length(src_type);
-			memcpy(dest, srcdata, *resultlen);
+		case SYBTIME:
+		case SYBDATE:
+		case SYB5BIGDATETIME:
+		case SYB5BIGTIME:
+			*resultlen = tds_get_size_by_type(src_type);
+			if (*resultlen > 0)
+				memcpy(dest, srcdata, *resultlen);
 			ret = CS_SUCCEED;
 			break;
 
@@ -682,6 +650,7 @@ cs_convert(CS_CONTEXT * ctx, CS_DATAFMT * srcfmt, CS_VOID * srcdata, CS_DATAFMT
 		case SYBDECIMAL:
 			src_len = tds_numeric_bytes_per_prec[((TDS_NUMERIC *) srcdata)->precision] + 2;
 		case SYBBITN:
+		case SYBUNIQUE:
 			memcpy(dest, srcdata, minlen);
 			*resultlen = minlen;
 
@@ -772,55 +741,26 @@ cs_convert(CS_CONTEXT * ctx, CS_DATAFMT * srcfmt, CS_VOID * srcdata, CS_DATAFMT
 	case SYBBITN:
 		/* fall trough, act same way of TINYINT */
 	case SYBINT1:
-		memcpy(dest, &(cres.ti), 1);
-		*resultlen = 1;
-		ret = CS_SUCCEED;
-		break;
+	case SYBUINT1:
 	case SYBINT2:
-		memcpy(dest, &(cres.si), 2);
-		*resultlen = 2;
-		ret = CS_SUCCEED;
-		break;
+	case SYBUINT2:
 	case SYBINT4:
-		memcpy(dest, &(cres.i), 4);
-		*resultlen = 4;
-		ret = CS_SUCCEED;
-		break;
+	case SYBUINT4:
 	case SYBINT8:
-		memcpy(dest, &(cres.bi), 8);
-		*resultlen = 8;
-		ret = CS_SUCCEED;
-		break;
+	case SYBUINT8:
 	case SYBFLT8:
-		memcpy(dest, &(cres.f), 8);
-		*resultlen = 8;
-		ret = CS_SUCCEED;
-		break;
 	case SYBREAL:
-		memcpy(dest, &(cres.r), 4);
-		*resultlen = 4;
-		ret = CS_SUCCEED;
-		break;
 	case SYBMONEY:
-
-		tdsdump_log(TDS_DBG_FUNC, "cs_convert() copying %d bytes to src\n", (int) sizeof(TDS_MONEY));
-		memcpy(dest, &(cres.m), sizeof(TDS_MONEY));
-		*resultlen = sizeof(TDS_MONEY);
-		ret = CS_SUCCEED;
-		break;
 	case SYBMONEY4:
-		memcpy(dest, &(cres.m4), sizeof(TDS_MONEY4));
-		*resultlen = sizeof(TDS_MONEY4);
-		ret = CS_SUCCEED;
-		break;
 	case SYBDATETIME:
-		memcpy(dest, &(cres.dt), sizeof(TDS_DATETIME));
-		*resultlen = sizeof(TDS_DATETIME);
-		ret = CS_SUCCEED;
-		break;
 	case SYBDATETIME4:
-		memcpy(dest, &(cres.dt4), sizeof(TDS_DATETIME4));
-		*resultlen = sizeof(TDS_DATETIME4);
+	case SYBTIME:
+	case SYBDATE:
+	case SYBUNIQUE:
+	case SYB5BIGDATETIME:
+	case SYB5BIGTIME:
+		*resultlen = tds_get_size_by_type(desttype);
+		memcpy(dest, &(cres.ti), *resultlen);
 		ret = CS_SUCCEED;
 		break;
 	case SYBNUMERIC:
@@ -894,37 +834,71 @@ cs_convert(CS_CONTEXT * ctx, CS_DATAFMT * srcfmt, CS_VOID * srcdata, CS_DATAFMT
 }
 
 CS_RETCODE
-cs_dt_crack(CS_CONTEXT * ctx, CS_INT datetype, CS_VOID * dateval, CS_DATEREC * daterec)
+cs_dt_crack_v2(CS_CONTEXT * ctx, CS_INT datetype, CS_VOID * dateval, CS_DATEREC * daterec)
 {
-	TDS_DATETIME *dt;
-	TDS_DATETIME4 *dt4;
 	TDSDATEREC dr;
+	TDS_INT tds_type;
+	bool extended = false;
 
-	tdsdump_log(TDS_DBG_FUNC, "cs_dt_crack(%p, %d, %p, %p)\n", ctx, datetype, dateval, daterec);
+	tdsdump_log(TDS_DBG_FUNC, "cs_dt_crack_v2(%p, %d, %p, %p)\n", ctx, datetype, dateval, daterec);
 
-	if (datetype == CS_DATETIME_TYPE) {
-		dt = (TDS_DATETIME *) dateval;
-		tds_datecrack(SYBDATETIME, dt, &dr);
-	} else if (datetype == CS_DATETIME4_TYPE) {
-		dt4 = (TDS_DATETIME4 *) dateval;
-		tds_datecrack(SYBDATETIME4, dt4, &dr);
-	} else {
+	switch (datetype) {
+	case CS_DATETIME_TYPE:
+		tds_type = SYBDATETIME;
+		break;
+	case CS_DATETIME4_TYPE:
+		tds_type = SYBDATETIME4;
+		break;
+	case CS_DATE_TYPE:
+		tds_type = SYBDATE;
+		break;
+	case CS_TIME_TYPE:
+		tds_type = SYBTIME;
+		break;
+	case CS_BIGDATETIME_TYPE:
+		tds_type = SYB5BIGDATETIME;
+		extended = true;
+		break;
+	case CS_BIGTIME_TYPE:
+		tds_type = SYB5BIGTIME;
+		extended = true;
+		break;
+	default:
 		return CS_FAIL;
 	}
-	daterec->dateyear = dr.year;
-	daterec->datemonth = dr.month;
-	daterec->datedmonth = dr.day;
-	daterec->datedyear = dr.dayofyear;
-	daterec->datedweek = dr.weekday;
+	tds_datecrack(tds_type, dateval, &dr);
+
+	/* Sybase CT-Library does not set these fields for CS_BIGTIME_TYPE */
+	if (tds_type != SYB5BIGTIME) {
+		daterec->dateyear = dr.year;
+		daterec->datemonth = dr.month;
+		daterec->datedmonth = dr.day;
+		daterec->datedyear = dr.dayofyear;
+		daterec->datedweek = dr.weekday;
+	}
 	daterec->datehour = dr.hour;
 	daterec->dateminute = dr.minute;
 	daterec->datesecond = dr.second;
-	daterec->datemsecond = dr.millisecond;
+	daterec->datemsecond = dr.decimicrosecond / 10000u;
 	daterec->datetzone = 0;
+	if (extended) {
+		daterec->datesecfrac = dr.decimicrosecond / 10u;
+		daterec->datesecprec = 1000000;
+	}
 
 	return CS_SUCCEED;
 }
 
+CS_RETCODE
+cs_dt_crack(CS_CONTEXT * ctx, CS_INT datetype, CS_VOID * dateval, CS_DATEREC * daterec)
+{
+	tdsdump_log(TDS_DBG_FUNC, "cs_dt_crack(%p, %d, %p, %p)\n", ctx, datetype, dateval, daterec);
+
+	if (datetype != CS_BIGDATETIME_TYPE && datetype != CS_BIGTIME_TYPE)
+		return cs_dt_crack_v2(ctx, datetype, dateval, daterec);
+	return CS_FAIL;
+}
+
 CS_RETCODE
 cs_loc_alloc(CS_CONTEXT * ctx, CS_LOCALE ** locptr)
 {
@@ -974,12 +948,10 @@ cs_locale(CS_CONTEXT * ctx, CS_INT action, CS_LOCALE * locale, CS_INT type, CS_V
 			}
 			
 			free(locale->charset);
-			locale->charset = (char *)malloc(buflen + 1);
+			locale->charset = tds_strndup(buffer, buflen);
 			if (!locale->charset)
 				break;
 
-			strncpy(locale->charset, (char *)buffer, buflen);
-			locale->charset[buflen] = '\0';
 			code = CS_SUCCEED;
 			break;
 
@@ -989,12 +961,10 @@ cs_locale(CS_CONTEXT * ctx, CS_INT action, CS_LOCALE * locale, CS_INT type, CS_V
 			}
 			
 			free(locale->language);
-			locale->language = (char *)malloc(buflen + 1);
+			locale->language = tds_strndup(buffer, buflen);
 			if (!locale->language)
 				break;
 
-			strncpy(locale->language, (char *)buffer, buflen);
-			locale->language[buflen] = '\0';
 			code = CS_SUCCEED;
 			break;
 
@@ -1019,21 +989,15 @@ cs_locale(CS_CONTEXT * ctx, CS_INT action, CS_LOCALE * locale, CS_INT type, CS_V
 			}
 			if (i) {
 				free(locale->language);
-				locale->language = (char *)malloc(i + 1);
+				locale->language = tds_strndup(b, i);
 				if (!locale->language)
 					break;
-
-				strncpy(locale->language, b, i);
-				locale->language[i] = '\0';
 			}
 			if (i != (buflen - 1)) {
 				free(locale->charset);
-				locale->charset = (char *)malloc(buflen - i);
+				locale->charset = tds_strndup(b + i + 1, buflen - i - 1);
 				if (!locale->charset)
 					break;
-				
-				strncpy(locale->charset, b + i + 1, buflen - i - 1);
-				locale->charset[buflen - i - 1] = '\0';
 			}
 			code = CS_SUCCEED;
 			break;
@@ -1046,12 +1010,10 @@ cs_locale(CS_CONTEXT * ctx, CS_INT action, CS_LOCALE * locale, CS_INT type, CS_V
 			}
 			
 			free(locale->collate);
-			locale->collate = (char *)malloc(buflen + 1);
+			locale->collate = tds_strndup(buffer, buflen);
 			if (!locale->collate)
 				break;
 
-			strncpy(locale->collate, (char *)buffer, buflen);
-			locale->collate[buflen] = '\0';
 			code = CS_SUCCEED;
 			break;
 		*/
@@ -1158,7 +1120,7 @@ cs_strbuild(CS_CONTEXT * ctx, CS_CHAR * buffer, CS_INT buflen, CS_INT * resultle
 	    CS_CHAR * formats, CS_INT formatlen, ...)
 {
 	va_list ap;
-	int rc;
+	TDSRET rc;
 
 	tdsdump_log(TDS_DBG_FUNC, "cs_strbuild(%p, %p, %d, %p, %p, %d, %p, %d)\n", 
 				ctx, buffer, buflen, resultlen, text, textlen, formats, formatlen);
@@ -1167,7 +1129,7 @@ cs_strbuild(CS_CONTEXT * ctx, CS_CHAR * buffer, CS_INT buflen, CS_INT * resultle
 	rc = tds_vstrbuild(buffer, buflen, resultlen, text, textlen, formats, formatlen, ap);
 	va_end(ap);
 
-	return rc == TDS_SUCCEED ? CS_SUCCEED : CS_FAIL;
+	return TDS_SUCCEED(rc) ? CS_SUCCEED : CS_FAIL;
 }
 
 CS_RETCODE
@@ -1358,12 +1320,12 @@ cs_diag_storemsg(CS_CONTEXT *context, CS_CLIENTMSG *message)
 		return CS_FAIL;
 	}
 
-	*curptr = (struct cs_diag_msg *) malloc(sizeof(struct cs_diag_msg));
+	*curptr = tds_new(struct cs_diag_msg, 1);
 	if (*curptr == NULL) { 
 		return CS_FAIL;
 	} else {
 		(*curptr)->next = NULL;
-		(*curptr)->msg  = malloc(sizeof(CS_CLIENTMSG));
+		(*curptr)->msg  = tds_new(CS_CLIENTMSG, 1);
 		if ((*curptr)->msg == NULL) {
 			return CS_FAIL;
 		} else {
